searchData={"items":[{"type":"module","title":"Peeper","doc":"`Peeper.GenServer` is an almost drop-in replacement for `GenServer` that preserves the state\n  between crashes. All the callbacks from `GenServer` are supported.\n\nInternally it creates a sub-supervision tree with an actual worker that delegates to the\n  implementation and a state keeper. That said, it creates three processes instead of one,\n  and this should be considered when building a very concurrent applications.\n\nThe main use-case would be a long-lived `GenServer` process with no error path handling at all\n  (yeah, that famous _Let It Crash_.) Using this abstraction, one is free to send unexpected\n  messages to the process, raise from its `handle_×××` clauses and whatnot.\n\nThere are two differencies compared to bare `GenServer`. `init/1` callback cannot return anything\n  but `{:ok, state}` or `{:ok, new_state, timeout() | :hibernate | {:continue, term()}` tuples\n  (this might have changed in the future,) and `Peeper.{call/3,cast/2,send/2}` is to be used\n  instead of `GenServer.{call/3,cast/2}` and `Kernel.send/2` (this is not gonna change.)\n\nPlease note, that whatever is set in `init/1` callback as a state, will be overriden by\n  the latest state available upon restarts. That being said, `init/1` would play its role\n  in setting the state during the first run only. \n\nInstead of using `Peeper`’s counterparts, one might either name the process and use `Name.GenServer`\n  as a name of the underlying `GenServer` or get the `GenServer`’s `pid` via `Peeper.gen_server/1`\n  and use `GenServer.{call/3,cast/2}` with it.","ref":"Peeper.html"},{"type":"module","title":"Example - Peeper","doc":"iex> {:ok, pid} = Peeper.Impls.Full.start_link(state: 0, name: Counter)\n    ...> Peeper.call(pid, :state)\n    0\n    iex> Peeper.cast(pid, :inc)\n    :ok\n    iex> GenServer.call(Peeper.gen_server(Counter), :state)\n    1\n    iex> Peeper.call(pid, :state)\n    1\n    iex> # emulate crash\n    ...> Process.exit(Peeper.Supervisor.worker(pid), :raise)\n    ...> %{} = Peeper.Supervisor.which_children(pid)\n    iex> Peeper.call(Counter, :state)\n    1\n    iex> Peeper.send(pid, :inc)\n    :inc\n    iex> Peeper.call(Counter, :state)\n    2\n\n### `start_link/1`\n\nThe function receives either an initial `state`, or a keyword having keys\n  `state` and (optionally) `name`. Also this keyword might have a configuration\n  for the top supervisor (keys: `[:strategy, :max_restarts, :max_seconds, :auto_shutdown]`.)\n\nAll other values passed would be re-passed to the underlying `GenServer`’s options.","ref":"Peeper.html#module-example"},{"type":"module","title":"Keeping ETS - Peeper","doc":"`Peeper` can preserve ETS tables created by the wrapped process between crashes in several ways.\n\nThe proper solution would be to add a `:heir` option to the ETS created by the process as\n\n```elixir\n:ets.new(name, [:private, :ordered_set, Peeper.heir(MyNamedPeeper)])\n```\n\nThat way the ETS will remain private and not readable by any other part of the system, although\n  it’ll be preserved between crashes _and_ might be transferred to another dynamic supervisor (see below.)\n\nIf for some reason setting `:heir` explicitly is not an option, one might pass \n  `keep_ets: true | :all | [ets_name()]` option to `Peeper.start_link/1`. It’s not efficient, because\n  ETS content will be passed tyo the backing state process every time the change to it happens to occur.\n\nIf the ETS created by the underlined process has `{:heir, other_process_pid()}` set, ithe behaviour\n  after a process crash is undefined, because ETS will be transferred to another process and reach out\n  of the control of `Peeper`.","ref":"Peeper.html#module-keeping-ets"},{"type":"module","title":"Moving to another DynamicSupervior - Peeper","doc":"The `Peeper` branch might be transferred an a whole to another dynamic supervisor in the following way\n\n```elixir\nPeeper.transfer(MyNamedPeeper, source_dynamic_supervisor, target_dynamic_supervisor)\n```\n\nThe `target_dynamic_supervisor` might be a remote `pid`, in such a case `Peeper` must be compiled \n  and loaded on the target node.","ref":"Peeper.html#module-moving-to-another-dynamicsupervior"},{"type":"module","title":"Listener - Peeper","doc":"One might pass `listener: MyListener` key to `PeeperImpl.start_link/1` where `MyListener`\n  should be an implementation of `Peeper.Listener` behaviour. The callbacks will be called\n  when the state of the underlying behaviour is _changed_ and on termination respectively.\n\nThat might be a good place to attach telemetry events, or logs, or whatnot.","ref":"Peeper.html#module-listener"},{"type":"function","title":"Peeper.call/3","doc":"The counterpart for `GenServer.call/3`. Uses the very same semantics.","ref":"Peeper.html#call/3"},{"type":"function","title":"Peeper.cast/3","doc":"The counterpart for `GenServer.cast/2`. Uses the very same semantics.","ref":"Peeper.html#cast/3"},{"type":"function","title":"Peeper.child_spec/1","doc":"Returns a specification to start a branch under a supervisor.","ref":"Peeper.html#child_spec/1"},{"type":"type","title":"Peeper.ets_option/0","doc":"The option accepted by `:ets.new/2`","ref":"Peeper.html#t:ets_option/0"},{"type":"function","title":"Peeper.gen_server/2","doc":"Returns the `pid` of the actual `GenServer`. Might be used to\navoid the necessity of calling other functions in this module.\n\nThe `pid` returned might be used directly in calls to\n`GenServer.{call/3,cast/2}` and/or `Kernel.send/2`","ref":"Peeper.html#gen_server/2"},{"type":"function","title":"Peeper.gen_server_name/1","doc":"Tries to produce a name for the underlying `GenServer` process.\nSucceeds if the name passed to `start_link/1` of the implementation\nis an `atom()`, `{:global, atom()}`, or `{:via, module(), atom()}`.\n\nIn such a case, the underlying `GenServer` module receives the name\n`Module.concat(name, GenServer)` and might be used as such.","ref":"Peeper.html#gen_server_name/1"},{"type":"function","title":"Peeper.heir/2","doc":"Returns a `{:heir, pid(), heir_data}` tuple where `pid` is the\npid of the state holder. Useful when the `GenServer`\nprocess creates persistent _ETS_, the result of this function is\nto be passed to _ETS_ options as a config parameter.\n\nIn that case, the tables will be given away to the state process\nand then retaken by the restarted `GenServer`.","ref":"Peeper.html#heir/2"},{"type":"type","title":"Peeper.name/0","doc":"The name of the processes branch","ref":"Peeper.html#t:name/0"},{"type":"function","title":"Peeper.send/3","doc":"The counterpart for `Kernel.send/2`. Uses the very same semantics.","ref":"Peeper.html#send/3"},{"type":"function","title":"Peeper.start_link/1","doc":"Starts a supervisor branch with the given options.","ref":"Peeper.html#start_link/1"},{"type":"function","title":"Peeper.transfer/4","doc":"Transfers the whole `Peeper` branch from one supervisor to another.","ref":"Peeper.html#transfer/4"},{"type":"behaviour","title":"Peeper.GenServer","doc":"A drop-in replacement for `use GenServer` that preserves state between crashes.","ref":"Peeper.GenServer.html"},{"type":"behaviour","title":"Overview - Peeper.GenServer","doc":"`Peeper.GenServer` enables you to create GenServer-like processes that maintain their state\neven when they crash. It accomplishes this through a specialized supervision tree that includes:\n\n1. A worker process (implementing your callbacks)\n2. A state keeper process (that preserves state between crashes)\n3. A supervisor process (managing the relationship between the worker and state keeper)","ref":"Peeper.GenServer.html#module-overview"},{"type":"behaviour","title":"Comparison with Standard GenServer - Peeper.GenServer","doc":"`Peeper.GenServer` supports all the standard GenServer callbacks with nearly identical\nsemantics, but with two important differences:\n\n1. The `init/1` callback can only return one of:\n   * `{:ok, state}`\n   * `{:ok, state, timeout | :hibernate | {:continue, term()}}`\n\n2. You must use the `Peeper` module's communication functions:\n   * `Peeper.call/3` instead of `GenServer.call/3`\n   * `Peeper.cast/2` instead of `GenServer.cast/2`\n   * `Peeper.send/2` instead of direct process messaging","ref":"Peeper.GenServer.html#module-comparison-with-standard-genserver"},{"type":"behaviour","title":"State Preservation Mechanism - Peeper.GenServer","doc":"When the worker process crashes, the supervisor restarts it, and the state keeper\nprovides the most recent state to the restarted worker. This makes the `init/1` callback\nfunction primarily important only during the first start - subsequent restarts will\noverride whatever state is set in `init/1` with the preserved state.","ref":"Peeper.GenServer.html#module-state-preservation-mechanism"},{"type":"behaviour","title":"Performance Considerations - Peeper.GenServer","doc":"Since `Peeper.GenServer` creates three processes instead of one, it introduces some\noverhead compared to a standard GenServer. Use it when state preservation between crashes\noutweighs the need for absolute performance in highly concurrent applications.","ref":"Peeper.GenServer.html#module-performance-considerations"},{"type":"behaviour","title":"Advanced Features - Peeper.GenServer","doc":"* **ETS Table Preservation**: Peeper can preserve ETS tables between crashes\n* **Process Dictionary Preservation**: Process dictionary entries are also preserved\n* **Listeners**: Attach listeners to monitor state changes and termination events\n* **Dynamic Supervisor Transfer**: Move Peeper processes between supervisors","ref":"Peeper.GenServer.html#module-advanced-features"},{"type":"behaviour","title":"Usage - Peeper.GenServer","doc":"Use `Peeper.GenServer` by adding it to your module:\n\n```elixir\ndefmodule MyStatePreservingServer do\n  use Peeper.GenServer\n  \n  # Define your callbacks similar to GenServer\nend\n```\n\nSee the examples below for more detailed usage patterns.","ref":"Peeper.GenServer.html#module-usage"},{"type":"macro","title":"Peeper.GenServer.__using__/1","doc":"Declares a `Peeper.GenServer` behaviour, injects `start_link/1` function\nand the child spec.","ref":"Peeper.GenServer.html#__using__/1"},{"type":"macro","title":"Customization Options - Peeper.GenServer.__using__/1","doc":"When using `Peeper.GenServer`, you can pass options that will be applied as defaults \nin all `start_link/1` calls:\n\n* `listener`: A module implementing the `Peeper.Listener` behaviour\n* `keep_ets`: Whether to preserve ETS tables between crashes (`true`, `:all`, or a list of table names)\n* Any other options that would be valid for `start_link/1`","ref":"Peeper.GenServer.html#__using__/1-customization-options"},{"type":"macro","title":"What's Injected - Peeper.GenServer.__using__/1","doc":"Using this macro injects the following functions into your module:\n\n* `start_link/1`: For starting the Peeper process\n* `child_spec/1`: For use in supervision trees\n* `stop/3`: For stopping the Peeper process","ref":"Peeper.GenServer.html#__using__/1-what-s-injected"},{"type":"macro","title":"Basic Example - Peeper.GenServer.__using__/1","doc":"```elixir\ndefmodule Counter do\n  use Peeper.GenServer\n\n  @impl Peeper.GenServer\n  def init(_) do\n    {:ok, 0}  # Initial state is 0\n  end\n\n  @impl Peeper.GenServer\n  def handle_call(:get, _from, state) do\n    {:reply, state, state}\n  end\n\n  @impl Peeper.GenServer\n  def handle_cast(:inc, state) do\n    {:noreply, state + 1}\n  end\n  \n  @impl Peeper.GenServer\n  def handle_cast({:add, n}, state) do\n    {:noreply, state + n}\n  end\nend\n```","ref":"Peeper.GenServer.html#__using__/1-basic-example"},{"type":"macro","title":"Example with Listener - Peeper.GenServer.__using__/1","doc":"```elixir\ndefmodule MyListener do\n  @behaviour Peeper.Listener\n  \n  @impl Peeper.Listener\n  def on_state_changed(old_state, new_state) do\n    # Log, send telemetry event, etc.\n    IO.puts(\"State changed from #{inspect(old_state)} to #{inspect(new_state)}\")\n    :ok\n  end\n  \n  @impl Peeper.Listener\n  def on_terminate(reason, final_state) do\n    IO.puts(\"Process terminated with reason: #{inspect(reason)}\")\n    IO.puts(\"Final state was: #{inspect(final_state)}\")\n    :ok\n  end\nend\n\ndefmodule MyGenServer do\n  use Peeper.GenServer, listener: MyListener\n  \n  @impl Peeper.GenServer\n  def init(args) do\n    {:ok, args}\n  end\n  \n  # ... other callbacks ...\nend\n```","ref":"Peeper.GenServer.html#__using__/1-example-with-listener"},{"type":"macro","title":"Example with ETS Table Preservation - Peeper.GenServer.__using__/1","doc":"```elixir\ndefmodule CacheServer do\n  use Peeper.GenServer, keep_ets: true\n  \n  @impl Peeper.GenServer\n  def init(_) do\n    # Create an ETS table that will be preserved across crashes\n    :ets.new(:cache, [:named_table, :set, :public])\n    {:ok, %{last_updated: nil}}\n  end\n  \n  @impl Peeper.GenServer\n  def handle_call({:get, key}, _from, state) do\n    result = :ets.lookup(:cache, key)\n    {:reply, result, state}\n  end\n  \n  @impl Peeper.GenServer\n  def handle_cast({:set, key, value}, state) do\n    :ets.insert(:cache, {key, value})\n    {:noreply, %{state | last_updated: System.monotonic_time()}}\n  end\n  \n  # Even if this process crashes, the ETS table will be preserved\nend\n```","ref":"Peeper.GenServer.html#__using__/1-example-with-ets-table-preservation"},{"type":"callback","title":"Peeper.GenServer.code_change/3","doc":"_See:_ `c:GenServer.code_change/3`","ref":"Peeper.GenServer.html#c:code_change/3"},{"type":"callback","title":"Peeper.GenServer.handle_call/3","doc":"_See:_ `c:GenServer.handle_call/3`","ref":"Peeper.GenServer.html#c:handle_call/3"},{"type":"callback","title":"Peeper.GenServer.handle_cast/2","doc":"_See:_ `c:GenServer.handle_cast/2`","ref":"Peeper.GenServer.html#c:handle_cast/2"},{"type":"callback","title":"Peeper.GenServer.handle_continue/2","doc":"_See:_ `c:GenServer.handle_continue/2`","ref":"Peeper.GenServer.html#c:handle_continue/2"},{"type":"callback","title":"Peeper.GenServer.handle_info/2","doc":"_See:_ `c:GenServer.handle_info/2`","ref":"Peeper.GenServer.html#c:handle_info/2"},{"type":"callback","title":"Peeper.GenServer.init/1","doc":"Initializes the GenServer state.\n\nThis callback serves a similar purpose to `c:GenServer.init/1` but has more restricted\nreturn values. Unlike standard GenServer's init/1, in Peeper.GenServer:\n\n* It may only return `{:ok, state}` or `{:ok, state, timeout | :hibernate | {:continue, term()}}`\n* It can't return `:ignore` or `{:stop, reason}`\n* It can't use other GenServer init/1 patterns like `{:ok, state, {:continue, term()}}`","ref":"Peeper.GenServer.html#c:init/1"},{"type":"callback","title":"State Preservation Behavior - Peeper.GenServer.init/1","doc":"It's important to understand that:\n\n1. During the first start, the state you set in init/1 becomes the initial state\n2. During restarts after crashes, the state from init/1 will be immediately overridden\n   by the state preserved in the Peeper state keeper process\n3. The continued_arg in `{:continue, continued_arg}` is still respected on all starts","ref":"Peeper.GenServer.html#c:init/1-state-preservation-behavior"},{"type":"callback","title":"Examples - Peeper.GenServer.init/1","doc":"```elixir\n# Basic initialization\n@impl Peeper.GenServer\ndef init(args) do\n  state = %{counter: 0, config: Keyword.get(args, :config, %{})}\n  {:ok, state}\nend\n\n# Initialization with continue\n@impl Peeper.GenServer\ndef init(args) do\n  {:ok, args, {:continue, :load_initial_data}}\nend\n```\n\n_See:_ `c:GenServer.init/1` for more details on the standard GenServer behavior.","ref":"Peeper.GenServer.html#c:init/1-examples"},{"type":"function","title":"Peeper.GenServer.start_link/2","doc":"Starts a `Peeper` sub-supervision process tree linked to the current process.\n\nThis is the underlying implementation function called by the module-specific `start_link/1`\nthat gets injected when using `use Peeper.GenServer`. It properly formats the options\nand delegates to `Peeper.Supervisor.start_link/1`.","ref":"Peeper.GenServer.html#start_link/2"},{"type":"function","title":"Parameters - Peeper.GenServer.start_link/2","doc":"* `module`: The module that implements the `Peeper.GenServer` behaviour\n* `opts`: Options for initializing and configuring the Peeper supervision tree","ref":"Peeper.GenServer.html#start_link/2-parameters"},{"type":"function","title":"Options - Peeper.GenServer.start_link/2","doc":"* Any non-keyword argument is treated as the initial state\n* If a keyword, it may include:\n  * `state`: The initial state value\n  * `listener`: Module implementing the `Peeper.Listener` behaviour\n  * Any other valid options for `Peeper.Supervisor.start_link/1`","ref":"Peeper.GenServer.html#start_link/2-options"},{"type":"function","title":"Returns - Peeper.GenServer.start_link/2","doc":"* `{:ok, pid}` if the server is started successfully\n* `{:error, reason}` if the server failed to start","ref":"Peeper.GenServer.html#start_link/2-returns"},{"type":"callback","title":"Peeper.GenServer.terminate/2","doc":"_See:_ `c:GenServer.terminate/2`","ref":"Peeper.GenServer.html#c:terminate/2"},{"type":"module","title":"Peeper.State","doc":"Core state management module for the Peeper framework.\n\nThis module is responsible for preserving and restoring state between crashes of\nthe wrapped GenServer. It acts as a private state keeper within the Peeper supervision\ntree, storing the latest state of the GenServer implementation.","ref":"Peeper.State.html"},{"type":"module","title":"Role in Supervision Tree - Peeper.State","doc":"`Peeper.State` is a child process in the Peeper supervision tree, started directly \nunder the main `Peeper.Supervisor`. It maintains:\n\n1. The current state of the wrapped GenServer\n2. Any ETS tables created by the process that need to be preserved\n3. The process dictionary content that needs to be restored after crashes","ref":"Peeper.State.html#module-role-in-supervision-tree"},{"type":"module","title":"Internal Functionality - Peeper.State","doc":"When the worker GenServer crashes and is restarted by the supervisor, it\nretrieves its last known state from this state keeper process. This allows\nthe GenServer to maintain state continuity despite crashes, implementing the\n\"Let It Crash\" philosophy without losing valuable state.\n\nThis module also handles:\n- ETS table ownership transfers between the worker and state processes\n- Process migration between supervisors (including remote nodes)\n- Storage of process dictionary data\n\nThis is an internal module that shouldn't be used directly by client code.\nInstead, interact with Peeper through the public API in the `Peeper` module.","ref":"Peeper.State.html#module-internal-functionality"},{"type":"function","title":"Peeper.State.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Peeper.State.html#child_spec/1"},{"type":"type","title":"Peeper.State.ets_dump/0","doc":"Represents a serialized ETS table dump consisting of:\n- The table name or identifier\n- A list of table options (access, heir, etc.)\n- The actual table data as a list of tuples\n\nThis format allows ETS tables to be recreated with identical configuration\nand data after a worker process crash.","ref":"Peeper.State.html#t:ets_dump/0"},{"type":"function","title":"Peeper.State.init/1","doc":"Initializes the State GenServer with the provided state information.\n\nThis function:\n1. Extracts the supervisor PID and any initial state data\n2. Sets up the internal ETS table to store state, dictionary, and ETS tables data\n3. Processes any inherited ETS tables that need special handling\n\nThe internal ETS table structure contains three key-value pairs:\n- `{:state, state}` - The actual application state to be preserved\n- `{:dictionary, dictionary}` - Process dictionary entries to be restored\n- `{:ets, ets}` - ETS tables information for recreation after a crash","ref":"Peeper.State.html#init/1"},{"type":"function","title":"Peeper.State.start_link/1","doc":"Starts a State GenServer process linked to the current process.\n\nThis function is called by the Peeper.Supervisor to initialize the state keeper\nthat will preserve state between worker restarts.","ref":"Peeper.State.html#start_link/1"},{"type":"function","title":"Parameters - Peeper.State.start_link/1","doc":"- `state`: A keyword list containing initialization parameters with the following keys:\n  - `:supervisor`: PID of the parent supervisor (required)\n  - `:state`: The initial state to be stored (optional)\n  - `:ets`: ETS tables to be preserved (optional)\n  - `:dictionary`: Process dictionary content to preserve (optional)","ref":"Peeper.State.html#start_link/1-parameters"},{"type":"module","title":"Peeper.Supervisor","doc":"Supervisor for the Peeper framework that manages the state-preserving GenServer processes.","ref":"Peeper.Supervisor.html"},{"type":"module","title":"Overview - Peeper.Supervisor","doc":"The `Peeper.Supervisor` is a key component in the Peeper architecture, responsible for:\n\n1. Creating and supervising a supervision tree for each Peeper.GenServer instance\n2. Managing the relationship between the worker process and state keeper process\n3. Ensuring state is preserved between crashes\n4. Handling process transfers between supervisors","ref":"Peeper.Supervisor.html#module-overview"},{"type":"module","title":"Supervision Strategy - Peeper.Supervisor","doc":"By default, `Peeper.Supervisor` uses a `:one_for_one` supervision strategy with the following configuration:\n\n* `max_restarts`: 1,000 (number of restarts allowed in a time period)\n* `max_seconds`: 10 (the time period for max_restarts)\n* `auto_shutdown`: `:never` (supervisor won't shut down automatically when a child terminates)\n\nThis strategy allows individual worker processes to crash and be restarted independently,\nwhile the state keeper process remains alive to provide the preserved state on restart.","ref":"Peeper.Supervisor.html#module-supervision-strategy"},{"type":"module","title":"Children - Peeper.Supervisor","doc":"The supervisor manages two worker processes:\n\n1. `Peeper.State` - Preserves the state between crashes\n2. `Peeper.Worker` - The actual GenServer implementation that delegates to your module","ref":"Peeper.Supervisor.html#module-children"},{"type":"module","title":"State Preservation Mechanism - Peeper.Supervisor","doc":"When a worker process crashes:\n\n1. The supervisor detects the crash and initiates a restart\n2. The new worker process starts and requests its previous state from the state keeper\n3. The state keeper provides the preserved state\n4. The worker process resumes operation with the preserved state\n\nThis handshake between the worker and state keeper happens transparently and enables \nthe \"Let It Crash\" philosophy without losing valuable state.","ref":"Peeper.Supervisor.html#module-state-preservation-mechanism"},{"type":"module","title":"Configuration - Peeper.Supervisor","doc":"When starting a Peeper.Supervisor, you can provide the following configuration options:\n\n* `strategy`: The restart strategy (default: `:one_for_one`)\n* `max_restarts`: Maximum number of restarts allowed in a time period (default: 1,000)\n* `max_seconds`: The time period for max_restarts (default: 10)\n* `auto_shutdown`: Supervisor shutdown behavior when a child terminates (default: `:never`)","ref":"Peeper.Supervisor.html#module-configuration"},{"type":"module","title":"Internal Architecture - Peeper.Supervisor","doc":"```\nPeeper.Supervisor (one_for_one)\n├── Peeper.State (preserves state)\n└── Peeper.Worker (runs user callbacks)\n    └── User's Peeper.GenServer implementation\n```\n\nThis modular architecture allows for:\n* Independent restart of workers without losing state\n* Clean separation of concerns\n* ETS table preservation between crashes\n* Process dictionary persistence","ref":"Peeper.Supervisor.html#module-internal-architecture"},{"type":"function","title":"Peeper.Supervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Peeper.Supervisor.html#child_spec/1"},{"type":"function","title":"Peeper.Supervisor.start_link/1","doc":"Starts a supervisor linked to the current process.","ref":"Peeper.Supervisor.html#start_link/1"},{"type":"function","title":"Options - Peeper.Supervisor.start_link/1","doc":"* `name`: Name to register the supervisor (required)\n* `impl`: The module implementing Peeper.GenServer callbacks (required)\n* `state`: Initial state for the GenServer (required)\n* `strategy`, `max_restarts`, `max_seconds`, `auto_shutdown`: \n    Standard supervisor configuration options\n* Any other options are passed to the Worker process","ref":"Peeper.Supervisor.html#start_link/1-options"},{"type":"function","title":"Returns - Peeper.Supervisor.start_link/1","doc":"* `{:ok, pid}` if the supervisor is started successfully\n* `{:error, {:already_started, pid}}` if the supervisor is already started\n* `{:error, term}` if supervision setup fails","ref":"Peeper.Supervisor.html#start_link/1-returns"},{"type":"function","title":"Peeper.Supervisor.state/2","doc":"Gets the PID of the state keeper process for the given Peeper supervisor.","ref":"Peeper.Supervisor.html#state/2"},{"type":"function","title":"Parameters - Peeper.Supervisor.state/2","doc":"* `pid`: PID or registered name of the Peeper supervisor\n* `delay`: Optional delay in milliseconds before attempting to get the state keeper PID","ref":"Peeper.Supervisor.html#state/2-parameters"},{"type":"function","title":"Returns - Peeper.Supervisor.state/2","doc":"PID of the state keeper process","ref":"Peeper.Supervisor.html#state/2-returns"},{"type":"function","title":"Peeper.Supervisor.whereis/1","doc":"Returns a local or remote pid if the `GenServer` is alive, `nil` otherwise","ref":"Peeper.Supervisor.html#whereis/1"},{"type":"function","title":"Peeper.Supervisor.which_children/2","doc":"Lists the children, awaiting for all to be restarted successfully.","ref":"Peeper.Supervisor.html#which_children/2"},{"type":"function","title":"Peeper.Supervisor.worker/2","doc":"Gets the PID of the worker process for the given Peeper supervisor.\n\nThis function is useful when you need to interact directly with the worker GenServer\nprocess using standard GenServer functions, bypassing Peeper's wrappers.","ref":"Peeper.Supervisor.html#worker/2"},{"type":"function","title":"Parameters - Peeper.Supervisor.worker/2","doc":"* `pid`: PID or registered name of the Peeper supervisor\n* `delay`: Optional delay in milliseconds before attempting to get the worker PID","ref":"Peeper.Supervisor.html#worker/2-parameters"},{"type":"function","title":"Returns - Peeper.Supervisor.worker/2","doc":"PID of the worker process","ref":"Peeper.Supervisor.html#worker/2-returns"},{"type":"function","title":"Example - Peeper.Supervisor.worker/2","doc":"```elixir\n# Get the worker pid and use it directly with GenServer\nworker_pid = Peeper.Supervisor.worker(my_peeper_process)\nGenServer.call(worker_pid, :message)\n```","ref":"Peeper.Supervisor.html#worker/2-example"},{"type":"behaviour","title":"Peeper.Listener","doc":"The listener that might be implemented and attached to the `Peeper.GenServer`\n  to receive notifications about process state changes and/or terminations.","ref":"Peeper.Listener.html"},{"type":"callback","title":"Peeper.Listener.on_state_changed/2","doc":"Called when the underlying `Peeper.GenServer`’s state gets changed","ref":"Peeper.Listener.html#c:on_state_changed/2"},{"type":"callback","title":"Peeper.Listener.on_terminate/2","doc":"Called when the underlying `Peeper.GenServer` gets terminated","ref":"Peeper.Listener.html#c:on_terminate/2"},{"type":"extras","title":"Peeper","doc":"# Peeper\n\n**Almost drop-in replacement for `GenServer` that preserves state between crashes.** **State Preservation**: Maintain process state even when crashes occur\n* **ETS Table Preservation**: Keep ETS tables alive between process restarts\n* **Process Dictionary Preservation**: Preserve process dictionary entries between crashes\n* **GenServer Compatible**: Use familiar GenServer callbacks and patterns\n* **Supervision Tree Integration**: Works within existing supervision hierarchies\n* **Listener Hooks**: Attach listeners to monitor state changes and termination events\n* **Dynamic Supervisor Transfer**: Move Peeper processes between supervisors (even across nodes)","ref":"readme.html"},{"type":"extras","title":"How It Works - Peeper","doc":"`Peeper.GenServer` is an almost drop-in replacement for `GenServer` that preserves state between crashes. All the callbacks from `GenServer` are supported.\n\nInternally, Peeper creates a specialized supervision tree with:\n\n1. A worker process (implementing your callbacks, similar to a regular GenServer)\n2. A state keeper process (that preserves state between crashes)\n3. A supervisor process (managing the relationship between the worker and state keeper)\n\nThis architecture allows Peeper to capture and restore state automatically when the worker process crashes, making it ideal for long-lived processes where you want to embrace the _\"Let It Crash\"_ philosophy without losing valuable state.","ref":"readme.html#how-it-works"},{"type":"extras","title":"Comparison with Standard GenServer - Peeper","doc":"Peeper maintains compatibility with GenServer while adding state persistence. There are two main differences:\n\n1. **Restricted `init/1` callback**: It can only return `{:ok, state}` or `{:ok, state, timeout | :hibernate | {:continue, term()}}` tuples.\n\n2. **Communication functions**: You must use:\n   * `Peeper.call/3` instead of `GenServer.call/3`\n   * `Peeper.cast/2` instead of `GenServer.cast/2`\n   * `Peeper.send/2` instead of `Kernel.send/2`\n\n> **Note**: Whatever is set in `init/1` callback will be overridden by the preserved state upon restarts. The `init/1` callback sets the state only during the first run.","ref":"readme.html#comparison-with-standard-genserver"},{"type":"extras","title":"Alternative Communication - Peeper","doc":"If you prefer standard GenServer communication functions, you can:\n* Name the process and use `Name.GenServer` as the name of the underlying GenServer\n* Get the worker's PID via `Peeper.gen_server/1` and use standard GenServer functions with it","ref":"readme.html#alternative-communication"},{"type":"extras","title":"Installation - Peeper","doc":"1. Add `peeper` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:peeper, \"~> 0.3.0\"}\n  ]\nend\n```\n\n2. Install the dependency:\n\n```bash\nmix deps.get\n```","ref":"readme.html#installation"},{"type":"extras","title":"Basic Usage - Peeper","doc":"","ref":"readme.html#basic-usage"},{"type":"extras","title":"Creating a Peeper.GenServer - Peeper","doc":"```elixir\ndefmodule Counter do\n  use Peeper.GenServer\n  \n  @impl Peeper.GenServer\n  def init(_) do\n    {:ok, 0}  # Initial state is 0\n  end\n  \n  @impl Peeper.GenServer\n  def handle_call(:get, _from, state) do\n    {:reply, state, state}\n  end\n  \n  @impl Peeper.GenServer\n  def handle_cast(:inc, state) do\n    {:noreply, state + 1}\n  end\nend\n```","ref":"readme.html#creating-a-peeper-genserver"},{"type":"extras","title":"Starting and Using - Peeper","doc":"```elixir\n# Start the server\n{:ok, pid} = Counter.start_link(name: CounterServer)\n\n# Use the server\nPeeper.call(pid, :get)  # Returns 0\nPeeper.cast(pid, :inc)  # Increments counter\nPeeper.call(CounterServer, :get)  # Returns 1\n\n# Crash the server\nProcess.exit(Peeper.gen_server(pid), :kill)\n\n# State is preserved after crash!\nPeeper.call(CounterServer, :get)  # Still returns 1\n```","ref":"readme.html#starting-and-using"},{"type":"extras","title":"Interactive Example - Peeper","doc":"```elixir\n# Start with initial state 0 and registered name Counter\niex> {:ok, pid} = Peeper.Impls.Full.start_link(state: this_is_pid())\n\n# Check the state\niex> Peeper.call(pid, :state)\n0\n\n# Update the state\niex> Peeper.cast(pid, :inc)\n:ok\n\n# State was updated\niex> Peeper.call(pid, :state)\n1\n\n# Emulate crash\niex> Process.exit(Peeper.Supervisor.worker(pid), :raise)\niex> %{} = Peeper.Supervisor.which_children(pid)\n\n# State is preserved after crash\niex> Peeper.call(pid, :state)\n1\n\n# Continue using the server\niex> Peeper.send(pid, :inc)\n:inc\niex> Peeper.call(pid, :state)\n2\n```","ref":"readme.html#interactive-example"},{"type":"extras","title":"Configuration Options - Peeper","doc":"When starting a Peeper GenServer, you can provide various configuration options:","ref":"readme.html#configuration-options"},{"type":"extras","title":"Basic Options - Peeper","doc":"* `state`: The initial state (required if not a keyword list)\n* `name`: Optional name to register the process (atom, {:global, term} or {:via, module, term})","ref":"readme.html#basic-options"},{"type":"extras","title":"ETS Table Options - Peeper","doc":"* `keep_ets`: Options for preserving ETS tables:\n  * `true`: Preserve all ETS tables owned by the process\n  * `:all`: Same as `true`\n  * `[table_name1, table_name2, ...]`: List of specific ETS table names to preserve","ref":"readme.html#ets-table-options"},{"type":"extras","title":"Listener Options - Peeper","doc":"* `listener`: A module implementing the `Peeper.Listener` behaviour","ref":"readme.html#listener-options"},{"type":"extras","title":"Supervisor Configuration - Peeper","doc":"* `strategy`: The restart strategy (default: `:one_for_all`)\n* `max_restarts`: Maximum number of restarts allowed in a time period (default: 3)\n* `max_seconds`: The time period for max_restarts (default: 5)\n* `auto_shutdown`: When a child process terminates, whether to automatically terminate the supervisor and all other children (default: `:never`)","ref":"readme.html#supervisor-configuration"},{"type":"extras","title":"Advanced Usage - Peeper","doc":"","ref":"readme.html#advanced-usage"},{"type":"extras","title":"Working with ETS Tables - Peeper","doc":"There are two ways to preserve ETS tables between crashes:\n\n#### Method 1: Using `Peeper.heir/2` (Recommended)\n\n```elixir\ndefmodule CacheServer do\n  use Peeper.GenServer\n  \n  @impl Peeper.GenServer\n  def init(_) do\n    # Create an ETS table with Peeper as heir\n    :ets.new(:my_cache, [:named_table, :set, :private, Peeper.heir(self())])\n    {:ok, %{last_access: nil}}\n  end\n  \n  # Rest of implementation...\nend\n```\n\nThis method is more efficient because the table ownership is transferred directly via the heir mechanism.\n\n#### Method 2: Using `keep_ets` Option\n\n```elixir\ndefmodule CacheServer do\n  use Peeper.GenServer, keep_ets: true  # Can also be :all or a list of table names\n  \n  @impl Peeper.GenServer\n  def init(_) do\n    :ets.new(:my_cache, [:named_table, :set])\n    {:ok, %{last_access: nil}}\n  end\n  \n  # Rest of implementation...\nend\n```\n\nThis method copies the table contents to the state keeper, so it's less efficient for large tables or frequent updates.","ref":"readme.html#working-with-ets-tables"},{"type":"extras","title":"Using Listeners - Peeper","doc":"Listeners can monitor state changes and termination events, useful for logging, telemetry, or other side effects:\n\n```elixir\ndefmodule MyListener do\n  @behaviour Peeper.Listener\n  \n  @impl Peeper.Listener\n  def on_state_changed(old_state, new_state) do\n    # Log state change, send metrics, etc.\n    Logger.info(\"State changed from #{inspect(old_state)} to #{inspect(new_state)}\")\n    :ok\n  end\n  \n  @impl Peeper.Listener\n  def on_terminate(reason, final_state) do\n    # Log termination\n    Logger.info(\"Process terminated with reason: #{inspect(reason)}\")\n    :ok\n  end\nend\n\ndefmodule MyServer do\n  use Peeper.GenServer, listener: MyListener\n  \n  # Implementation...\nend\n```","ref":"readme.html#using-listeners"},{"type":"extras","title":"Transferring Between Supervisors - Peeper","doc":"Peeper allows transferring a process between different dynamic supervisors, even across nodes:\n\n```elixir\n# Start source and target supervisors\n{:ok, source_sup} = DynamicSupervisor.start_link(strategy: :one_for_one)\n{:ok, target_sup} = DynamicSupervisor.start_link(strategy: :one_for_one)\n\n# Start a Peeper process under the source supervisor\n{:ok, peeper_pid} = DynamicSupervisor.start_child(\n  source_sup,\n  {MyPeeperServer, state: initial_state, name: MyServer}\n)\n\n# Later, transfer the process to the target supervisor\nPeeper.transfer(MyServer, source_sup, target_sup)\n\n# The process continues running under the target supervisor with preserved state\n```","ref":"readme.html#transferring-between-supervisors"},{"type":"extras","title":"Best Practices - Peeper","doc":"","ref":"readme.html#best-practices"},{"type":"extras","title":"When to Use Peeper - Peeper","doc":"Peeper is ideal for:\n* Long-running processes where state preservation is critical\n* Scenarios where you want to embrace \"Let It Crash\" while keeping state\n* Services that use ETS tables for caching or state storage\n* Applications where process recovery should be transparent to clients\n\nNot recommended for:\n* Very high-throughput, performance-critical processes (due to the extra process overhead)\n* Cases where state loss on crash is acceptable or desirable\n* Short-lived or ephemeral processes","ref":"readme.html#when-to-use-peeper"},{"type":"extras","title":"Supervision Strategy - Peeper","doc":"Peeper works best with the following supervision pattern:\n\n```\nApplication Supervisor (one_for_one)\n├── Other Workers/Supervisors\n├── DynamicSupervisor\n│   └── Peeper Supervision Tree\n│       ├── Peeper.Supervisor\n│       ├── Peeper.State (state keeper)\n│       └── Your Peeper.GenServer (worker)\n```\n\nThis isolation ensures that Peeper process crashes don't affect other parts of your application.","ref":"readme.html#supervision-strategy"},{"type":"extras","title":"Handling ETS Tables - Peeper","doc":"* Prefer the `Peeper.heir/2` approach for ETS tables when possible\n* Be cautious when setting custom heirs other than Peeper on ETS tables\n* For very large tables, consider using a separate persistent storage mechanism","ref":"readme.html#handling-ets-tables"},{"type":"extras","title":"State Size Considerations - Peeper","doc":"* Keep your GenServer state reasonably sized - large states impact performance\n* Consider using ETS tables (with Peeper's heir mechanism) for larger datasets\n* For very large states, consider external persistence (database, disk)\n* Avoid storing resources that can't be serialized (PIDs, ports, references) in state\n* Be mindful of how state size affects message passing between processes","ref":"readme.html#state-size-considerations"},{"type":"extras","title":"Message Handling - Peeper","doc":"* Implement proper timeouts in your handle_call callbacks\n* Use handle_continue for breaking down complex operations\n* Handle unexpected messages in handle_info to prevent crashes from stray messages","ref":"readme.html#message-handling"},{"type":"extras","title":"Performance Considerations - Peeper","doc":"","ref":"readme.html#performance-considerations"},{"type":"extras","title":"Process Overhead - Peeper","doc":"Peeper creates three processes for each `Peeper.GenServer` (compared to one for a standard GenServer):\n\n1. The supervisor process\n2. The state keeper process\n3. The worker process (your actual GenServer implementation)\n\nThis increases the memory footprint and adds some communication overhead. The impact is typically negligible for most applications but can become significant with hundreds or thousands of Peeper processes.","ref":"readme.html#process-overhead"},{"type":"extras","title":"State Transfer Cost - Peeper","doc":"When the state changes, Peeper must transfer it to the state keeper process. This introduces a small overhead proportional to the state size. For extremely frequent state changes (thousands per second), this can impact performance.","ref":"readme.html#state-transfer-cost"},{"type":"extras","title":"ETS Table Considerations - Peeper","doc":"When using the `keep_ets: true` approach, Peeper copies the entire ETS table content on each state update. For large or frequently updated tables, this can be expensive.\n\nUsing the `Peeper.heir/2` approach is much more efficient as it only transfers table ownership, not content.","ref":"readme.html#ets-table-considerations"},{"type":"extras","title":"Benchmarks - Peeper","doc":"Some rough performance comparisons:\n\n| Operation | Standard GenServer | Peeper.GenServer |\n|-----------|-------------------|-----------------|\n| Start | ~1x | ~1.5-2x slower |\n| call/cast with small state | ~1x | ~1.1x slower |\n| call/cast with large state | ~1x | ~1.5-2x slower |\n| Process memory | ~1x | ~3x more |\n\nFor most applications, this performance difference is not significant, but it's worth considering for high-throughput services.","ref":"readme.html#benchmarks"},{"type":"extras","title":"Troubleshooting - Peeper","doc":"","ref":"readme.html#troubleshooting"},{"type":"extras","title":"Common Issues and Solutions - Peeper","doc":"#### Process Not Responding After Crash\n\n**Problem**: After a crash, your Peeper process doesn't respond to messages.\n\n**Possible causes**:\n* Supervisor restart limit exceeded\n* Invalid state that causes immediate crash on restart\n\n**Solutions**:\n* Check logs for crash reason\n* Increase restart limits in supervisor configuration\n* Add defensive code to handle invalid states\n\n#### State Not Being Preserved\n\n**Problem**: After a crash, your Peeper process starts with initial state instead of preserved state.\n\n**Possible causes**:\n* State keeper process crashed\n* Supervisor was restarted completely\n\n**Solutions**:\n* Make sure you're using correct supervision strategy\n* Check if the state keeper process is alive with `Peeper.Supervisor.state(pid)`\n\n#### ETS Tables Not Being Preserved\n\n**Problem**: ETS tables disappear after a crash despite using `keep_ets` or `Peeper.heir/2`.\n\n**Possible causes**:\n* Using the wrong table name\n* Custom heir settings overriding Peeper's heir\n* Table is not owned by the Peeper worker process\n\n**Solutions**:\n* Verify table ownership with `:ets.info(table_name, :owner)`\n* Check that you're not setting another heir after creating the table\n\n#### Slow Performance\n\n**Problem**: Your Peeper implementation is slow compared to a standard GenServer.\n\n**Possible causes**:\n* Large state size\n* Frequent state updates\n* Many ETS tables with `keep_ets: true`\n\n**Solutions**:\n* Optimize state size\n* Use `Peeper.heir/2` for ETS tables\n* Consider if state preservation is necessary for this specific process","ref":"readme.html#common-issues-and-solutions"},{"type":"extras","title":"License - Peeper","doc":"MIT License - see the [LICENSE](LICENSE) file for details.","ref":"readme.html#license"},{"type":"extras","title":"Documentation - Peeper","doc":"For more detailed documentation, visit [HexDocs](https://hexdocs.pm/peeper).","ref":"readme.html#documentation"},{"type":"extras","title":"Contributing - Peeper","doc":"Contributions are welcome! Please feel free to submit a Pull Request.","ref":"readme.html#contributing"},{"type":"extras","title":"Changelog","doc":"# Changelog\n\nAll notable changes to this project will be documented in this file.\n\nThe format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),\nand this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).\n\n## [Unreleased]","ref":"changelog.html"},{"type":"extras","title":"Added - Changelog","doc":"- Comprehensive property-based tests for state preservation\n- Enhanced documentation for all modules with detailed examples\n- Improved test organization with describe blocks and better coverage\n- New benchmarking capabilities to measure performance overhead\n- Additional edge case tests for supervisor restart limits","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"- Updated dependencies to latest versions\n- Improved project structure with better organization of test files\n- Enhanced README with clearer installation and usage instructions\n\n## [0.3.0] - 2023-03-22","ref":"changelog.html#changed"},{"type":"extras","title":"Added - Changelog","doc":"- Support for process migration between supervisors, including remote nodes\n- Process dictionary preservation between crashes\n- Listener hooks for state changes and termination events\n- ETS table ownership transfer capabilities\n- More examples in the documentation","ref":"changelog.html#added-1"},{"type":"extras","title":"Changed - Changelog","doc":"- Improved supervision tree structure for better fault tolerance\n- Enhanced API for interacting with Peeper processes\n- Refactored internal state handling for better performance\n- Updated documentation with more comprehensive examples\n- Refined error handling and logging","ref":"changelog.html#changed-1"},{"type":"extras","title":"Fixed - Changelog","doc":"- Issue with ETS table inheritance during crashes\n- Race condition in state keeper process\n- Improved handling of edge cases in process initialization\n\n## [0.2.0] - 2023-01-15","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"- Support for ETS table preservation between crashes\n- Handle_continue callback support in GenServer implementation\n- More comprehensive typespecs for better Dialyzer checks\n- Auto-shutdown configuration options for supervisors","ref":"changelog.html#added-2"},{"type":"extras","title":"Changed - Changelog","doc":"- Refined API for better ergonomics\n- Improved documentation with more complete examples\n- Enhanced test suite with better coverage","ref":"changelog.html#changed-2"},{"type":"extras","title":"Fixed - Changelog","doc":"- Issue with state restoration after complex crashes\n- Bug in child_spec generation\n- Race condition in process restart sequence\n\n## [0.1.0] - 2022-11-10","ref":"changelog.html#fixed-1"},{"type":"extras","title":"Added - Changelog","doc":"- Initial release with core functionality\n- Basic GenServer-like API with state preservation between crashes\n- Simple supervision tree for managing worker and state processes\n- Support for named processes\n- Basic documentation and examples\n- Initial test suite covering basic functionality\n\n[Unreleased]: https://github.com/am-kantox/peeper/compare/v0.3.0...HEAD\n[0.3.0]: https://github.com/am-kantox/peeper/compare/v0.2.0...v0.3.0\n[0.2.0]: https://github.com/am-kantox/peeper/compare/v0.1.0...v0.2.0\n[0.1.0]: https://github.com/am-kantox/peeper/releases/tag/v0.1.0","ref":"changelog.html#added-3"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.1"}}